#!python3

"""
    Configuration for project documentation using Sphinx.
"""

# standard
import sys
from datetime import date, datetime
from importlib.metadata import version
from os import environ, path

import toml

__version__ = version("qgis-plugin-ci")
pyproject = toml.load("../pyproject.toml")

# -- Build environment -----------------------------------------------------
on_rtd = environ.get("READTHEDOCS", None) == "True"

# -- Project information -----------------------------------------------------
author = ", ".join([a["name"] for a in pyproject["project"]["authors"]])
__copyright__ = f"2019 - {date.today().year}, {author}"
project = "QGIS Plugin CI"
version = release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # Sphinx included
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    # 3rd party
    "myst_parser",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_sitemap",
]


# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = {".md": "markdown", ".rst": "restructuredtext"}
autosectionlabel_prefix_document = True
# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", ".venv", "Thumbs.db", ".DS_Store", "_output", "ext_libs"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

# -- Theme

# final URL
html_baseurl = pyproject["project"]["urls"]["homepage"]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["static"]
html_extra_path = ["robots.txt"]

# theme
html_theme = "furo"
html_theme_options = {
    "navigation_with_keys": True,
    "source_repository": pyproject["project"]["urls"]["repository"],
    "source_branch": "main",
    "source_directory": "docs/",
}


# -- EXTENSIONS --------------------------------------------------------

myst_enable_extensions = [
    "colon_fence",
    "deflist",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "substitution",
]

myst_heading_anchors = 3

# replacement variables
myst_substitutions = {
    "author": author,
    "date_update": datetime.now().strftime("%d %B %Y"),
    "description": pyproject["project"]["description"],
    "repo_url": pyproject["project"]["urls"]["homepage"],
    "title": project,
    "version": version,
}

myst_url_schemes = ("http", "https", "mailto")

# Configuration for intersphinx (refer to others docs).
intersphinx_mapping = {"python": ("https://docs.python.org/3/", None)}

# sitemap
sitemap_url_scheme = "{link}"

# -- Options for Sphinx API doc ----------------------------------------------


# run api doc
def run_apidoc(_):
    from sphinx.ext.apidoc import main

    cur_dir = path.normpath(path.dirname(__file__))
    output_path = path.join(cur_dir, "_apidoc")
    modules = path.normpath(path.join(cur_dir, "../qgispluginci"))
    exclusions = ["../.venv", "../scripts", "../server", "../tests"]
    main(["-e", "-f", "-M", "-o", output_path, modules] + exclusions)


# launch setup
def setup(app):
    app.connect("builder-inited", run_apidoc)



import os


def can_skip_test():
    """Check when a test can run."""
    is_ci = os.getenv("CI") == "true"
    is_main_repo = os.getenv("GITHUB_REPOSITORY") == "opengisch/qgis-plugin-ci"
    if is_ci and is_main_repo:
        # Always run the test on CI
        return False

    if not os.getenv("tx_api_token") or not os.getenv("github_token"):
        # On local, the token must be set
        return True

    return False



#! /usr/bin/env python

# standard library
import logging
import os
import unittest

# 3rd party
import yaml

# project
from qgispluginci.parameters import Parameters
from qgispluginci.translation import Translation

# Tests
from .utils import can_skip_test

# Logging
logger = logging.getLogger(__name__)

# Ensuring proper ordering for tests sensitive to state
unittest.TestLoader.sortTestMethodsUsing = None


class TestTranslation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Initialize the test case"""
        with open(".qgis-plugin-ci") as f:
            arg_dict = yaml.safe_load(f)
        cls.parameters = Parameters(arg_dict)
        cls.tx_api_token = os.getenv("tx_api_token")
        assert cls.tx_api_token

    def setUp(self):
        """Initialize the next test method (run before every test method)"""
        self.t = Translation(self.parameters, tx_api_token=self.tx_api_token)

    def tearDown(self):
        self.t.tx_client.delete_project()
        """
        try:
            self.t.tx_client.delete_team(f"{self.parameters.project_slug}-team")
        except PyTransifexException as error:
            logger.debug(error)
        """

    @unittest.skipIf(can_skip_test(), "Missing tx_api_token")
    def test_creation(self):
        """
        Translation initialized from setUp, so we 'fake' a new test
        by tearing it down
        """
        self.t = Translation(self.parameters, tx_api_token=self.tx_api_token)
        self.t.tx_client.delete_project()
        self.assertFalse(self.t.tx_client.project_exists(self.parameters.project_slug))
        self.t = Translation(self.parameters, tx_api_token=self.tx_api_token)
        self.assertTrue(self.t.tx_client.project_exists(self.parameters.project_slug))
        self.assertEqual(len(self.t.tx_client.list_resources()), 1)

    @unittest.skipIf(can_skip_test(), "Missing tx_api_token")
    def test_push(self):
        self.t.update_strings()
        self.t.push()

    @unittest.skipIf(can_skip_test(), "Missing tx_api_token")
    def test_pull(self):
        self.t.pull()
        self.t.compile_strings()


if __name__ == "__main__":
    unittest.main()



import logging
from pathlib import Path

import requests
from transifex.api import transifex_api as tx_api
from transifex.api.jsonapi.exceptions import DoesNotExist

from qgispluginci.translation_clients.baseclient import BaseClient, TranslationConfig

# GLOBALS
logger = logging.getLogger(__name__)


class TransifexClient(BaseClient):
    def __init__(
        self, config: TranslationConfig, update_string_fcn, create_project: bool = True
    ):
        super().__init__(config, update_string_fcn, create_project)

    def login(self):
        tx_api.setup(auth=self.config.api_token)
        self.organization = self.get_organization()
        logger.info(f"Logged in as organization: {self.config.organization_name}")

    def get_organization(self):
        return tx_api.Organization.get(slug=self.config.organization_name)

    def get_project(self):
        try:
            return (
                self.get_organization()
                .fetch("projects")
                .get(slug=self.config.project_slug)
            )
        except DoesNotExist:
            return None

    def project_exists(self, project_slug: str) -> bool:
        """Check if the project exists in the remote Transifex repository"""
        try:
            if self.get_project():
                return True
            return False
        except DoesNotExist:
            return False

    def create_project(self):
        source_language = tx_api.Language.get(code=self.config.source_language_code)
        project_name = self.config.project_name or self.config.project_slug

        if self.config.private:
            tx_api.Project.create(
                name=project_name,
                slug=self.config.project_slug,
                source_language=source_language,
                private=self.config.private,
                organization=self.get_organization(),
            )
        else:
            tx_api.Project.create(
                name=project_name,
                slug=self.config.project_slug,
                source_language=source_language,
                private=self.config.private,
                organization=self.get_organization(),
                repository_url=self.config.repository_url,
            )

        return self.get_project()

    def delete_project(self):
        project = self.get_project()
        if project:
            project.delete()

    def create_resource(self):
        resource = tx_api.Resource.create(
            project=self.get_project(),
            name=self.config.resource_slug,
            slug=self.config.resource_slug,
            i18n_format=tx_api.I18nFormat(id=self.config.i18n_type),
        )

        with open(self.config.resource_file_path) as fh:
            content = fh.read()

        tx_api.ResourceStringsAsyncUpload.upload(content, resource=resource)
        logger.info(f"Resource created: {self.config.resource_slug}")

    def get_resource(self):
        resources = self.get_project().fetch("resources")
        if not resources:
            return None
        return resources.get(slug=self.config.resource_slug)

    def list_resources(self):
        if resources := self.get_project().fetch("resources"):
            return list(resources.all())
        else:
            return []

    def list_languages(self):
        languages = self.get_project().fetch("languages").all()
        return [lang.code for lang in languages]

    def create_language(self, language_code: str, coordinators):
        if language := tx_api.Language.get(code=language_code):
            logger.debug(f"Adding {language.code} to {self.config.project_slug}")
            self.get_project().add("languages", [language])

        # if coordinators:
        #    self.get_project().add("coordinators", coordinators)

    def update_source_translation(self):
        with open(self.config.resource_file_path) as fh:
            content = fh.read()

        tx_api.ResourceStringsAsyncUpload.upload(content, resource=self.get_resource())
        logger.info(f"Source updated for resource: {self.config.resource_slug}")

    def get_translation(
        self,
        language_code: str,
        path_to_output_file: str,
    ) -> str:
        """Fetch the translation resource matching the given language"""
        path_to_parent = Path(path_to_output_file).parent

        Path.mkdir(path_to_parent, parents=True, exist_ok=True)
        language = tx_api.Language.get(code=language_code)

        url = tx_api.ResourceTranslationsAsyncDownload.download(
            resource=self.get_resource(), language=language
        )

        r = requests.get(url)
        # Transifex returns None encoding and the apparent_encoding is Windows-1254 what leads to malformed
        # result strings.
        # So we set the encoding hardcoded to utf-8.
        if not r.encoding:
            r.encoding = "utf-8"
        translated_content = r.text
        with open(path_to_output_file, "w") as fh:
            fh.write(translated_content)

        logger.info(f"Translations '{language_code}' downloaded")
        return str(path_to_output_file)



"""
/***************************************************************************

                                 QgisLocator

                             -------------------
        begin                : 2018-05-03
        copyright            : (C) 2018 by Denis Rouzaud
        email                : denis@opengis.ch
        git sha              : $Format:%H$
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""


def classFactory(iface):
    """Load plugin.

    :param iface: A QGIS interface instance.
    :type iface: QgsInterface
    """
    #
    from .qgis_plugin_ci_testing_plugin import QgisPluginCiTesting

    return QgisPluginCiTesting(iface)



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest test.test_changelog
        # for specific test
        python -m unittest test.test_changelog.TestChangelog.test_has_changelog
"""

# standard library
import tempfile
import unittest
from pathlib import Path

# project
from qgispluginci.changelog import ChangelogParser
from qgispluginci.version_note import VersionNote

# ############################################################################
# ########## Classes #############
# ################################


class TestChangelog(unittest.TestCase):
    def test_has_changelog(self):
        """Test changelog path logic."""
        # using this repository as parent folder
        self.assertTrue(ChangelogParser.has_changelog())
        self.assertIsInstance(ChangelogParser.CHANGELOG_FILEPATH, Path)

        # using the fixture subfolder as string
        self.assertTrue(ChangelogParser.has_changelog(parent_folder="test/fixtures"))
        self.assertIsInstance(ChangelogParser.CHANGELOG_FILEPATH, Path)

        # using the fixture subfolder as pathlib.Path
        self.assertTrue(
            ChangelogParser.has_changelog(parent_folder=Path("test/fixtures"))
        )
        self.assertIsInstance(ChangelogParser.CHANGELOG_FILEPATH, Path)

        # with a path to a file, must raise a type error
        with self.assertRaises(SystemExit):
            ChangelogParser.has_changelog(parent_folder=Path(__file__))
        self.assertIsNone(ChangelogParser.CHANGELOG_FILEPATH, None)

        # with a path to a folder which doesn't exist, must raise a file exists error
        with self.assertRaises(SystemExit):
            ChangelogParser.has_changelog(parent_folder=Path("imaginary_path"))

    def test_changelog_content(self):
        """Test version content from changelog."""
        # parser
        parser = ChangelogParser(parent_folder="test/fixtures")
        self.assertIsInstance(parser.CHANGELOG_FILEPATH, Path)

        # Unreleased doesn't count
        self.assertEqual(7, len(parser._parse()))

        # This version doesn't exist
        self.assertIsNone(parser.content("0.0.0"))

        expected_checks = {
            "10.1.0-beta1": (
                "- This is the latest documented version in this changelog\n"
                "- The changelog module is tested against these lines\n"
                "- Be careful modifying this file"
            ),
            "10.1.0-alpha1": (
                "- This is a version with a prerelease in this changelog\n"
                "- The changelog module is tested against these lines\n"
                "- Be careful modifying this file"
                # "\n" TODO Fixed section is missing
                # "- trying with a subsection in a version note"
            ),
            "10.0.1": "- End of year version",
            "10.0.0": "- A\n- B\n- C",
            "9.10.1": "- D\n- E\n- F",
            "v0.1.1": (
                '* Tag with a "v" prefix to check the regular expression\n'
                "* Previous version"
            ),
            "0.1.0": "* Very old version",
        }
        for version, expected in expected_checks.items():
            with self.subTest(i=version):
                self.assertEqual(parser.content(version), expected)

    def test_changelog_content_latest(self):
        """Test against the latest special option value. \
        See: https://github.com/opengisch/qgis-plugin-ci/pull/33
        """
        # expected result
        expected_latest = (
            "- This is the latest documented version in this changelog\n"
            "- The changelog module is tested against these lines\n"
            "- Be careful modifying this file"
        )

        # get latest
        parser = ChangelogParser(parent_folder="test/fixtures")
        self.assertEqual(expected_latest, parser.content("latest"))

        self.assertEqual("10.1.0-beta1", parser.latest_version())

    def test_changelog_content_ci_fake(self):
        """Test specific fake version used in tests."""
        parser = ChangelogParser()
        fake_version_content = parser.content(tag="0.1.2")

        # expected result
        expected = (
            "(This version note is used in unit-tests)\n\n"
            '* Tag without "v" prefix\n'
            "* Add a CHANGELOG.md file for testing"
        )

        self.assertIsInstance(fake_version_content, str)
        self.assertEqual(expected, fake_version_content)

    def test_different_changelog_file(self):
        """Test against a different changelog filename."""
        old = Path("test/fixtures/CHANGELOG.md")
        new_folder = Path(tempfile.mkdtemp())
        new_path = new_folder / Path("CHANGELOG-branch-X.md")
        self.assertFalse(new_path.exists())

        new_path.write_text(old.read_text())

        self.assertTrue(
            ChangelogParser.has_changelog(
                parent_folder=new_folder,
                changelog_path=new_path,
            )
        )

    def test_changelog_last_items(self):
        """Test last items from changelog."""
        # on fixture changelog
        parser = ChangelogParser(parent_folder="test/fixtures")
        last_items = parser.last_items(3)
        self.assertIsInstance(last_items, str)

        # on repository changelog
        parser = ChangelogParser()
        last_items = parser.last_items(3)
        self.assertIsInstance(last_items, str)

    def test_changelog_version_note(self):
        """Test version note named tuple structure and mechanisms."""
        # parser
        parser = ChangelogParser(parent_folder="test/fixtures")
        self.assertIsInstance(parser.CHANGELOG_FILEPATH, Path)

        # content parsed
        changelog_content = parser._parse()
        self.assertEqual(len(changelog_content), 7)

        # loop on versions
        for version in changelog_content:
            version_note = VersionNote(*version)
            self.assertIsInstance(version_note.date, str)
            self.assertTrue(hasattr(version_note, "is_prerelease"))
            self.assertTrue(hasattr(version_note, "version"))
            if len(version_note.prerelease):
                self.assertEqual(version_note.is_prerelease, True)

    def test_version_note_tuple(self):
        """Test the version note tuple."""
        parser = ChangelogParser(parent_folder="test/fixtures")

        version = parser._version_note("0.0.0")
        self.assertIsNone(version)

        version = parser._version_note("10.1.0-beta1")
        self.assertEqual("10", version.major)
        self.assertEqual("1", version.minor)
        self.assertEqual("0", version.patch)
        self.assertEqual("", version.url)
        self.assertEqual("beta1", version.prerelease)
        self.assertTrue(version.is_prerelease)
        self.assertEqual("", version.separator)  # Not sure what is the separator
        self.assertEqual("2021/02/08", version.date)
        self.assertEqual(
            (
                "- This is the latest documented version in this changelog\n"
                "- The changelog module is tested against these lines\n"
                "- Be careful modifying this file"
            ),
            version.text,
        )

        version = parser._version_note("10.0.1")
        self.assertEqual("", version.prerelease)
        self.assertFalse(version.is_prerelease)


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



from typing import NamedTuple


class VersionNote(NamedTuple):
    major: str = None
    minor: str = None
    patch: str = None
    url: str = None
    prerelease: str = None
    separator: str = None
    date: str = None
    text_raw: str = None

    @property
    def text(self) -> str:
        """Remove many \n at the start and end of the string."""
        return self.text_raw.strip()

    @property
    def is_prerelease(self) -> bool:
        if self.prerelease and len(self.prerelease):
            return True
        else:
            return False

    @property
    def version(self) -> str:
        if self.prerelease:
            return f"{self.major}.{self.minor}.{self.patch}-{self.prerelease}"
        else:
            return f"{self.major}.{self.minor}.{self.patch}"



#!/usr/bin/env python3
import argparse
import logging
from importlib.metadata import version

from qgispluginci.changelog import ChangelogParser
from qgispluginci.parameters import Parameters
from qgispluginci.release import release
from qgispluginci.translation import Translation

__version__ = version("qgis-plugin-ci")
__title__ = "QGISPluginCI"


def cli():
    # create the top-level parser
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # Optional verbosity counter (eg. -v, -vv, -vvv, etc.)
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=1,
        dest="verbosity",
        help="Verbosity level: None = WARNING, -v = INFO, -vv = DEBUG",
    )

    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
    )

    subparsers = parser.add_subparsers(
        title="commands", description="qgis-plugin-ci command", dest="command"
    )

    # package
    package_parser = subparsers.add_parser(
        "package", help="creates an archive of the plugin"
    )
    package_parser.add_argument("release_version", help="The version to be released")
    package_parser.add_argument(
        "--transifex-token",
        help="The Transifex API token. If specified translations will be pulled and compiled.",
    )
    package_parser.add_argument(
        "-u",
        "--plugin-repo-url",
        help="If specified, a XML repository file will be created in the current directory, the zip URL will use this parameter.",
    )
    package_parser.add_argument(
        "-c",
        "--allow-uncommitted-changes",
        action="store_true",
        help="If omitted, uncommitted changes are not allowed before packaging. If specified and some changes are "
        "detected, a hard reset on a stash create will be used to revert changes made by qgis-plugin-ci.",
    )
    package_parser.add_argument(
        "-d",
        "--disable-submodule-update",
        action="store_true",
        help="If omitted, a git submodule is updated. If specified, git submodules will not be updated/initialized before packaging.",
    )
    package_parser.add_argument(
        "-a",
        "--asset-path",
        action="append",
        help="An additional asset path to add. Can be specified multiple times.",
    )
    package_parser.add_argument(
        "--no-validation",
        action="store_true",
        help="Turn off validation of `release version`",
    )

    # changelog
    changelog_parser = subparsers.add_parser(
        "changelog", help="gets the changelog content"
    )
    changelog_parser.add_argument(
        "release_version",
        help="The version to be released. If nothing is speficied, \
                                      the latest version specified into the changelog is used.",
        default="latest",
    )

    # release
    release_parser = subparsers.add_parser("release", help="release the plugin")
    release_parser.add_argument(
        "release_version", help="The version to be released (x.y.z)."
    )
    release_parser.add_argument(
        "--release-tag",
        help="The release tag, if different from the version (e.g. vx.y.z).",
    )
    release_parser.add_argument(
        "--transifex-token",
        help="The Transifex API token. If specified translations will be pulled and compiled.",
    )
    release_parser.add_argument(
        "--github-token",
        help="The Github API token. If specified, the archive will be pushed to an already existing release.",
    )
    release_parser.add_argument(
        "-r",
        "--create-plugin-repo",
        action="store_true",
        help="Will create a XML repo as a Github release asset. Github token is required.",
    )
    release_parser.add_argument(
        "-c",
        "--allow-uncommitted-changes",
        action="store_true",
        help="If omitted, uncommitted changes are not allowed before releasing. If specified and some changes are "
        "detected, a hard reset on a stash create will be used to revert changes made by qgis-plugin-ci.",
    )
    release_parser.add_argument(
        "-d",
        "--disable-submodule-update",
        action="store_true",
        help="If omitted, a git submodule is updated. If specified, git submodules will not be updated/initialized before packaging.",
    )
    release_parser.add_argument(
        "-a",
        "--asset-path",
        action="append",
        help="An additional asset path to add. Can be specified multiple times.",
    )
    release_parser.add_argument(
        "--alternative-repo-url",
        help="The URL of the endpoint to publish the plugin (defaults to plugins.qgis.org)",
    )
    release_parser.add_argument(
        "--osgeo-username", help="The Osgeo user name to publish the plugin."
    )
    release_parser.add_argument(
        "--osgeo-password", help="The Osgeo password to publish the plugin."
    )

    # pull-translation
    pull_tr_parser = subparsers.add_parser(
        "pull-translation", help="pull translations from Transifex"
    )
    pull_tr_parser.add_argument("transifex_token", help="The Transifex API token")
    pull_tr_parser.add_argument(
        "--compile", action="store_true", help="Will compile TS files into QM files"
    )

    # push-translation
    push_tr_parser = subparsers.add_parser(
        "push-translation", help="update strings and push translations"
    )
    push_tr_parser.add_argument("transifex_token", help="The Transifex API token")

    args = parser.parse_args()
    Parameters.validate_args(args)

    # set log level depending on verbosity argument
    args.verbosity = 40 - (10 * args.verbosity) if args.verbosity > 0 else 0
    logging.basicConfig(
        level=args.verbosity,
        format="%(asctime)s||%(levelname)s||%(module)s||%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    console = logging.StreamHandler()
    console.setLevel(args.verbosity)

    # add the handler to the root logger
    logger = logging.getLogger(__title__)
    logger.debug(f"Log level set: {logging}")

    # if no command is passed, print the help and exit
    if not args.command:
        parser.print_help()
        parser.exit()

    exit_val = 0

    # Initialize Parameters
    parameters = Parameters.make_from(args=args)
    # CHANGELOG
    if args.command == "changelog":
        try:
            c = ChangelogParser(
                changelog_path=parameters.changelog_path,
            )
            content = c.content(args.release_version)
            if content:
                print(content)  # noqa: T2
        except Exception as exc:
            logger.error("Something went wrong reading the changelog.", exc_info=exc)

        return exit_val

    # PACKAGE
    if args.command == "package":
        release(
            parameters,
            release_version=args.release_version,
            tx_api_token=args.transifex_token,
            allow_uncommitted_changes=args.allow_uncommitted_changes,
            plugin_repo_url=args.plugin_repo_url,
            disable_submodule_update=args.disable_submodule_update,
            asset_paths=args.asset_path,
        )

    # RELEASE
    elif args.command == "release":
        release(
            parameters,
            release_version=args.release_version,
            release_tag=args.release_tag,
            tx_api_token=args.transifex_token,
            github_token=args.github_token,
            upload_plugin_repo_github=args.create_plugin_repo,
            alternative_repo_url=args.alternative_repo_url,
            osgeo_username=args.osgeo_username,
            osgeo_password=args.osgeo_password,
            allow_uncommitted_changes=args.allow_uncommitted_changes,
            disable_submodule_update=args.disable_submodule_update,
            asset_paths=args.asset_path,
        )

    # TRANSLATION PULL
    elif args.command == "pull-translation":
        t = Translation(parameters, args.transifex_token)
        t.pull()
        if args.compile:
            t.compile_strings()

    # TRANSLATION PUSH
    elif args.command == "push-translation":
        t = Translation(parameters, args.transifex_token)
        t.update_strings()
        t.push()

    return exit_val



import glob
import logging
import subprocess
import sys
from pathlib import Path

from qgispluginci.exceptions import (
    TransifexManyResources,
    TransifexNoResource,
    TranslationFailed,
)
from qgispluginci.parameters import Parameters
from qgispluginci.translation_clients.baseclient import TranslationConfig
from qgispluginci.translation_clients.transifex import TransifexClient
from qgispluginci.utils import touch_file

# GLOBALS
logger = logging.getLogger(__name__)


class Translation:
    def __init__(
        self, parameters: Parameters, tx_api_token: str, create_project: bool = True
    ):
        """
        Parameters
        ----------
        parameters:

        tx_api_token:
            Transifex API token

        create_project:
            if True, it will create the project, resource and language on Transifex

        """
        self.parameters = parameters

        plugin_path = self.parameters.plugin_path
        tx = self.parameters.transifex_resource
        lang = self.parameters.translation_source_language
        self.ts_file = f"{plugin_path}/i18n/{tx}_{lang}.ts"

        tx_config = TranslationConfig(
            api_token=tx_api_token,
            organization_name=parameters.transifex_organization,
            project_slug=parameters.transifex_project,
            repository_url=parameters.repository_url,
            source_language_code=parameters.translation_source_language,
            resource_file_path=self.ts_file,
            resource_slug=self.parameters.transifex_resource,
        )
        self.tx_client = TransifexClient(tx_config, self.update_strings, create_project)

    def update_strings(self):
        """
        Update TS files from plugin source strings
        """
        source_py_files = []
        source_ui_files = []
        relative_path = f"./{self.parameters.plugin_path}"
        for ext in ("py", "ui"):
            for file in glob.glob(
                f"{self.parameters.plugin_path}/**/*.{ext}",
                recursive=True,
            ):
                file_path = str(Path(file).relative_to(relative_path))
                if ext == "py":
                    source_py_files.append(file_path)
                else:
                    source_ui_files.append(file_path)

        touch_file(self.ts_file)

        project_file = Path(self.parameters.plugin_path).joinpath(
            self.parameters.plugin_name + ".pro"
        )

        with open(project_file, "w") as f:
            source_py_files = " ".join(source_py_files)
            source_ui_files = " ".join(source_ui_files)
            assert f.write("CODECFORTR = UTF-8\n")
            assert f.write(f"SOURCES = {source_py_files}\n")
            assert f.write(f"FORMS = {source_ui_files}\n")
            assert f.write(
                f"TRANSLATIONS = {Path(self.ts_file).relative_to(relative_path)}\n"
            )
            f.flush()
            f.close()

        cmd = [self.parameters.pylupdate5_path, "-noobsolete", str(project_file)]

        output = subprocess.run(cmd, capture_output=True, text=True)

        project_file.unlink()

        if output.returncode != 0:
            logger.error(
                f"Translation failed: {output.stderr}", exc_info=TranslationFailed()
            )
            sys.exit(1)
        else:
            logger.info(f"Successfully run pylupdate5: {output.stdout}")

    def compile_strings(self):
        """
        Compile TS file into QM files
        """
        cmd = [self.parameters.lrelease_path]
        for file in glob.glob(f"{self.parameters.plugin_path}/i18n/*.ts"):
            cmd.append(file)
        output = subprocess.run(cmd, capture_output=True, text=True)
        if output.returncode != 0:
            logger.error(
                f"Translation failed: {output.stderr}", exc_info=TranslationFailed()
            )
            sys.exit(1)
        else:
            logger.info(f"Successfully run lrelease: {output.stdout}")

    def pull(self):
        """
        Pull TS files from Transifex
        """
        resource = self.tx_client.get_resource()
        existing_langs = self.tx_client.list_languages()
        logger.info(
            f"{len(existing_langs)} languages found for resource {resource}:"
            f" ({existing_langs})"
        )
        for lang in self.parameters.translation_languages:
            if lang not in existing_langs:
                logger.debug(f"Creating missing language: {lang}")
                self.tx_client.create_language(
                    language_code=lang,
                    coordinators=[self.parameters.transifex_coordinator],
                )
                existing_langs.append(lang)
        for lang in existing_langs:
            ts_file = f"{self.parameters.plugin_path}/i18n/{self.parameters.transifex_resource}_{lang}.ts"
            logger.debug(f"Downloading translation file: {ts_file}")
            self.tx_client.get_translation(
                language_code=lang,
                path_to_output_file=ts_file,
            )

    def push(self):
        self.tx_client.get_resource()
        logger.debug(
            f"Pushing resource: {self.parameters.transifex_resource} "
            f"with file {self.ts_file}"
        )
        self.tx_client.update_source_translation()



#! python3  # noqa E265

"""
    Parameters management.
"""

# ############################################################################
# ########## Libraries #############
# ##################################

import configparser
import datetime
import logging
import os
import re
import sys

# standard library
from argparse import Namespace
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, Optional, Tuple

import toml
import yaml

# 3rd party
from slugify import slugify

from qgispluginci.exceptions import ConfigurationNotFound

# ############################################################################
# ########## Globals #############
# ################################


DASH_WARNING = "Dash in the plugin name is causing issues with QGIS plugin manager"

logger = logging.getLogger(__name__)


# ############################################################################
# ########## Classes #############
# ################################


class Parameters:
    """
    Attributes
    ----------
    plugin_path: str
        The directory of the source code in the repository.
        Defaults to: `slugify(plugin_name)`

    github_organization_slug: str
        The organization slug on SCM host (e.g. Github) and translation platform (e.g. Transifex).
        Not required when running on Travis since deduced from `$TRAVIS_REPO_SLUG`environment variable.

    project_slug: str
        The project slug on SCM host (e.g. Github) and translation platform (e.g. Transifex).
        Not required when running on Travis since deduced from `$TRAVIS_REPO_SLUG`environment variable.

    transifex_coordinator: str
        The username of the coordinator in Transifex.
        Required to create new languages.

    transifex_organization: str
        The organization name in Transifex
        Defaults to: the GitHub organization slug

    transifex_project: str
        The project on Transifex, which can be different from the one on GitHub.
        Defaults to: the project_slug

    transifex_resource: str
        The resource name in transifex
        Defaults to: the project_slug

    translation_source_language:
        The source language for translations.
        Defaults to: 'en'

    translation_languages:
        List of languages.

    changelog_include:
        If the changelog must be added when releasing a version AND if there is a CHANGELOG.md file
        Defaults to True

    changelog_path:
        Path to the CHANGELOG.md relative to the configuration file. Defaults to CHANGELOG.md

    changelog_number_of_entries:
        Number of changelog entries to add in the metdata.txt
        Defaults to 3

    create_date: datetime.date
        The date of creation of the plugin.
        The would be used in the custom repository XML.
        Format: YYYY-MM-DD

    lrelease_path: str
        The path of lrelease executable

    pylupdate5_path: str
        The path of pylupdate executable


    """

    @classmethod
    def make_from(
        cls, *, args: Optional[Any] = None, path_to_config_file: Optional[Path] = None
    ) -> "Parameters":
        """
        Instantiate from a config file or by exploring the filesystem
        Accepts an argparse Namespace for backward compatibility.
        """
        supported_config_file_names = {
            "setup.cfg",
            ".qgis-plugin-ci",
            "pyproject.toml",
        }
        configuration_not_found = ConfigurationNotFound(
            ".qgis-plugin-ci or setup.cfg or pyproject.toml with a 'qgis-plugin-ci' section have not been found."
        )

        def load_config(path_to_config_file: Path, file_name: str) -> Dict[str, Any]:
            if file_name == "setup.cfg":
                config = configparser.ConfigParser()
                config.read(path_to_config_file)
                arg_dict = dict(config.items("qgis-plugin-ci"))
            else:
                with open(path_to_config_file) as fh:
                    if file_name == ".qgis-plugin-ci":
                        arg_dict = yaml.safe_load(fh)
                    elif file_name == "pyproject.toml":
                        contents = toml.load(fh)
                        arg_dict = contents["tool"]["qgis-plugin-ci"]
                    else:
                        raise configuration_not_found
            return arg_dict

        def explore_config() -> Dict[str, Any]:
            for file_name in supported_config_file_names:
                path_to_file = Path(file_name)
                if path_to_file.is_file():
                    try:
                        return load_config(path_to_file, path_to_file.name)
                    except (
                        ConfigurationNotFound,
                        configparser.NoSectionError,
                        KeyError,
                    ):
                        pass
            raise configuration_not_found

        if path_to_config_file:
            file_name = path_to_config_file.name

            if not (
                path_to_config_file.is_file()
                and file_name in supported_config_file_names
            ):
                raise configuration_not_found
            config_dict = load_config(path_to_config_file, file_name)
        else:
            config_dict = explore_config()
        return cls(config_dict)

    def __init__(self, definition: Dict[str, Any]):
        self.plugin_path = definition.get("plugin_path")

        get_metadata = self.collect_metadata()
        self.plugin_name = get_metadata("name")
        self.plugin_slug = slugify(self.plugin_name)
        self.project_slug = definition.get(
            "project_slug",
            os.environ.get("TRAVIS_REPO_SLUG", f".../{self.plugin_slug}").split("/")[1],
        )
        self.github_organization_slug = definition.get(
            "github_organization_slug",
            os.environ.get("TRAVIS_REPO_SLUG", "").split("/")[0],
        )
        self.transifex_coordinator = definition.get("transifex_coordinator", "")
        self.transifex_organization = definition.get(
            "transifex_organization", self.github_organization_slug
        )
        self.translation_source_language = definition.get(
            "translation_source_language", "en"
        )
        self.translation_languages = definition.get("translation_languages", {})
        self.transifex_project = definition.get("transifex_project", self.project_slug)
        self.transifex_resource = definition.get(
            "transifex_resource", self.project_slug
        )
        self.create_date = datetime.datetime.strptime(
            str(definition.get("create_date", datetime.date.today())), "%Y-%m-%d"
        )
        self.lrelease_path = definition.get("lrelease_path", "lrelease")
        self.pylupdate5_path = definition.get("pylupdate5_path", "pylupdate5")
        changelog_include = definition.get("changelog_include", True)
        if isinstance(changelog_include, str):
            self.changelog_include = changelog_include.lower() in [
                "true",
                "1",
                "t",
                "y",
            ]
        else:
            self.changelog_include = changelog_include
        self.changelog_number_of_entries = definition.get(
            "changelog_number_of_entries", 3
        )
        self.changelog_path = definition.get("changelog_path", "CHANGELOG.md")

        # read from metadata
        if not self.plugin_path:
            # This tool can be used outside of a QGIS plugin to read a changelog file
            return

        self.author = get_metadata("author", "")
        self.description = get_metadata("description")
        self.qgis_minimum_version = get_metadata("qgisMinimumVersion")
        self.icon = get_metadata("icon", "")
        self.tags = get_metadata("tags", "")
        self.experimental = get_metadata("experimental", False)
        self.deprecated = get_metadata("deprecated", False)
        self.issue_tracker = get_metadata("tracker")
        self.homepage = get_metadata("homepage", "")
        if self.homepage == "":
            logger.warning(
                "Homepage is not given in the metadata. "
                "It is a mandatory information to publish "
                "the plugin on the QGIS official repository."
            )
        self.repository_url = get_metadata("repository")

    @staticmethod
    def get_release_version_patterns() -> Dict[str, re.Pattern]:
        return {
            "simple": r"\d+\.\d+$",
            "double": r"\d+\.\d+\.\d+$",
            "v2": r"^v\d+\.\d+$",
            "v3": r"^v\d+\.\d+\.\d+$",
            # See https://github.com/semver/semver/blob/master/semver.md#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
            "semver": r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$",
        }

    @staticmethod
    def validate_args(args: Namespace):
        """
        Raise an exception just in case:
        - the user didn't opt-out of validation using the `--no-validation` flag; and
        - the value of `release_version` matches no supported pattern.
        In any case, warn the user if the value of `release_version` doesn't match the semver pattern.
        """
        if not hasattr(args, "release_version") or not args.release_version:
            return

        patterns = Parameters.get_release_version_patterns()
        semver_compliance = re.match(patterns.pop("semver"), args.release_version)

        if not semver_compliance:
            logging.warning(
                f"Be aware that '{args.release_version}' is not a semver-compliant version."
            )

        if args.no_validation:
            logging.warning("Disabled release version validation.")
            return

        if semver_compliance or any(
            re.match(other_pattern, args.release_version)
            for other_pattern in patterns.values()
        ):
            return

        raise ValueError(
            f"""
            Unable to validate the release version '{args.release_version}'.
            Please use a release version identifier such as '1.0.1' (recommended, semantic versioning), 'v1.1.1', 'v1.1', or '1.1'.
            Otherwise you can disable validation by running this command again with an extra '--no-validation' flag.
            Semantic versioning (semvar) identifiers are recommended.
            Take a look at https://en.wikipedia.org/wiki/Software_versioning#Semantic_versioning for a refresher."
        """
        )

    @staticmethod
    def archive_name(
        plugin_name, release_version: str, experimental: bool = False
    ) -> str:
        """
        Returns the archive file name
        """
        # zipname: use dot before version number
        # and not dash since it's causing issues #22
        if "-" in plugin_name:
            logger.warning(DASH_WARNING)

        experimental = "-experimental" if experimental else ""
        return f"{plugin_name}{experimental}.{release_version}.zip"

    def collect_metadata(self) -> Callable[[str, Optional[Any]], Any]:
        """
        Returns a closure capturing a Dict of metadata, allowing to retrieve one
        value after the other while also iterating over the file once.
        """
        metadata_file = f"{self.plugin_path}/metadata.txt"
        metadata = {}
        with open(metadata_file) as fh:
            for line in fh:
                split = line.strip().split("=", 1)
                if len(split) == 2:
                    metadata[split[0]] = split[1]

        def get_metadata(key: str, default_value: Optional[Any] = None) -> Any:
            if not self.plugin_path:
                return ""

            value = metadata.get(key, None)
            if value:
                return value
            elif default_value is not None:
                return default_value
            else:
                logger.error(f"Mandatory key is missing in metadata: {key}")
                sys.exit(1)

        return get_metadata

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        """Allows to represent attributes as dict, list, etc."""
        for k in vars(self):
            yield k, self.__getattribute__(k)

    def __str__(self) -> str:
        """Allows to represent instances as a string."""
        return str(dict(self))


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    pass



from typing import NamedTuple


class TranslationConfig(NamedTuple):
    api_token: str
    organization_name: str
    project_slug: str
    resource_file_path: str
    resource_slug: str

    private: bool = False
    project_name: str = None
    i18n_type: str = "QT"
    repository_url: str = None
    source_language_code: str = "en"


class BaseClient:
    def __init__(
        self, config: TranslationConfig, update_string_fcn, create_project: bool = True
    ):
        """
        Parameters
        ----------
        config:
            The config for the Translation platform

        create_project:
            if True, it will create the project, resource and language on Transifex

        """
        self.config = config
        self.login()
        self.project = self.get_project()
        if not self.project and create_project:
            self.project = self.create_project()
            update_string_fcn()
            self.create_resource()

    def login(self):
        raise NotImplementedError

    def get_project(self):
        raise NotImplementedError

    def project_exists(self):
        raise NotImplementedError

    def create_project(self):
        raise NotImplementedError

    def delete_project(self):
        raise NotImplementedError

    def create_resource(self):
        raise NotImplementedError

    def list_resources(self):
        raise NotImplementedError

    def get_resource(self):
        raise NotImplementedError

    def list_languages(self):
        raise NotImplementedError

    def create_language(self, language_code: str):
        raise NotImplementedError

    def update_source_translation(self):
        raise NotImplementedError

    def get_translation(
        self,
        language_code: str,
        path_to_output_file: str,
    ) -> str:
        raise NotImplementedError



import logging
import os
import re
from math import floor
from math import log as math_log
from math import pow
from typing import Union

from qgispluginci.version_note import VersionNote

# GLOBALS
logger = logging.getLogger(__name__)


def replace_in_file(file_path: str, pattern, new: str, encoding: str = "utf8"):
    with open(file_path, encoding=encoding) as f:
        content = f.read()
    content = re.sub(pattern, new, content, flags=re.M)
    with open(file_path, "w", encoding=encoding) as f:
        f.write(content)


def configure_file(source_file: str, dest_file: str, replace: dict):
    with open(source_file, encoding="utf-8") as f:
        content = f.read()
    for pattern, new in replace.items():
        content = re.sub(pattern, new, content, flags=re.M)
    with open(dest_file, "w", encoding="utf-8") as f:
        f.write(content)


def convert_octets(octets: int) -> str:
    """Convert a mount of octets in readable size.

    :param int octets: mount of octets to convert

    :Example:

    .. code-block:: python

        >>> convert_octets(1024)
        "1ko"
    """
    # check zero
    if octets == 0:
        return "0 octet"

    # conversion
    size_name = ("octets", "Ko", "Mo", "Go", "To", "Po")
    i = int(floor(math_log(octets, 1024)))
    p = pow(1024, i)
    s = round(octets / p, 2)

    return f"{s} {size_name[i]}"


def touch_file(path, update_time: bool = False, create_dir: bool = True):
    basedir = os.path.dirname(path)
    if create_dir and not os.path.exists(basedir):
        os.makedirs(basedir)
    with open(path, "a"):
        if update_time:
            os.utime(path, None)
        else:
            pass


def parse_tag(version_tag: str) -> Union[VersionNote, None]:
    """Parse a tag and determine the semantic version."""
    components = version_tag.split("-")
    items = components[0].split(".")

    try:
        if len(components) == 2:
            return VersionNote(
                major=items[0], minor=items[1], patch=items[2], prerelease=components[1]
            )
        else:
            return VersionNote(major=items[0], minor=items[1], patch=items[2])
    except IndexError:
        return VersionNote()



#! /usr/bin/env python

# standard
import argparse
import filecmp
import os
import re
import unittest
import urllib.request
from itertools import product
from pathlib import Path
from tempfile import mkstemp
from zipfile import ZipFile

# 3rd party
import yaml
from github import Github, GithubException

# Project
from qgispluginci.changelog import ChangelogParser
from qgispluginci.exceptions import GithubReleaseNotFound
from qgispluginci.parameters import DASH_WARNING, Parameters
from qgispluginci.release import release
from qgispluginci.translation import Translation
from qgispluginci.utils import replace_in_file

# Tests
from .utils import can_skip_test

# If changed, also update CHANGELOG.md
RELEASE_VERSION_TEST = "0.1.2"


class TestRelease(unittest.TestCase):
    def setUp(self):
        self.setup_params = Parameters.make_from(
            path_to_config_file=Path("test/fixtures/setup.cfg")
        )
        self.qgis_plugin_config_params = Parameters.make_from(
            path_to_config_file=Path("test/fixtures/.qgis-plugin-ci")
        )
        self.pyproject_params = Parameters.make_from(
            path_to_config_file=Path("test/fixtures/pyproject.toml")
        )
        self.tx_api_token = os.getenv("tx_api_token")
        self.github_token = os.getenv("github_token")
        self.repo = None
        self.t = None
        if self.github_token:
            print("init Github")
            self.repo = Github(self.github_token).get_repo("opengisch/qgis-plugin-ci")
        self.clean_assets()

    def tearDown(self):
        self.clean_assets()

    def clean_assets(self):
        if self.repo:
            rel = None
            try:
                rel = self.repo.get_release(id=RELEASE_VERSION_TEST)
            except GithubException:
                raise GithubReleaseNotFound(f"Release {RELEASE_VERSION_TEST} not found")
            if rel:
                print("deleting release assets")
                for asset in rel.get_assets():
                    print(f"  delete {asset.name}")
                    asset.delete_asset()
        if self.t:
            self.t._t.delete_project(self.qgis_plugin_config_params.project_slug)

    def test_dict_from_config(self):
        with self.subTest():
            self.assertTrue(dict(self.qgis_plugin_config_params))
            self.assertTrue(dict(self.pyproject_params))
            self.assertTrue(dict(self.setup_params))

    def test_release_from_dot_qgis_plugin_ci(self):
        release(self.qgis_plugin_config_params, RELEASE_VERSION_TEST)

    def test_release_from_pyproject(self):
        print(self.pyproject_params)
        release(self.pyproject_params, RELEASE_VERSION_TEST)

    @unittest.skipIf(can_skip_test(), "Missing tx_api_token")
    def test_release_with_transifex(self):
        Translation(self.qgis_plugin_config_params, tx_api_token=self.tx_api_token)
        release(
            self.qgis_plugin_config_params,
            RELEASE_VERSION_TEST,
            tx_api_token=self.tx_api_token,
        )

    def test_zipname(self):
        """Tests about the zipname for the QGIS plugin manager.

        See #22 about dash
        and also capital letters
        """
        self.assertEqual(
            "my_plugin-experimental.0.0.0.zip",
            Parameters.archive_name("my_plugin", "0.0.0", True),
        )

        self.assertEqual(
            "My_Plugin.0.0.0.zip", Parameters.archive_name("My_Plugin", "0.0.0", False)
        )

        with self.assertLogs(
            logger="qgispluginci.parameters", level="WARNING"
        ) as captured:
            Parameters.archive_name("my-plugin", "0.0.0")
        self.assertEqual(
            len(captured.records), 1
        )  # check that there is only one log message
        self.assertEqual(captured.records[0].getMessage(), DASH_WARNING)

    @unittest.skipIf(can_skip_test(), "Missing github_token")
    def test_release_upload_github(self):
        release(
            self.qgis_plugin_config_params,
            RELEASE_VERSION_TEST,
            github_token=self.github_token,
            upload_plugin_repo_github=True,
        )

        # check the custom plugin repo
        _, xml_repo = mkstemp(suffix=".xml")
        url = f"https://github.com/opengisch/qgis-plugin-ci/releases/download/{RELEASE_VERSION_TEST}/plugins.xml"
        print(f"retrieve repo from {url}")
        urllib.request.urlretrieve(url, xml_repo)
        replace_in_file(
            xml_repo,
            r"<update_date>[\w-]+<\/update_date>",
            "<update_date>__TODAY__</update_date>",
        )
        if not filecmp.cmp("test/plugins.xml.expected", xml_repo, shallow=False):
            import difflib

            with open("test/plugins.xml.expected") as f:
                text1 = f.readlines()
            with open(xml_repo) as f:
                text2 = f.readlines()
            self.assertFalse(True, "\n".join(difflib.unified_diff(text1, text2)))

        # compare archive file size
        gh_release = self.repo.get_release(id=RELEASE_VERSION_TEST)
        archive_name = self.qgis_plugin_config_params.archive_name(
            self.qgis_plugin_config_params.plugin_path, RELEASE_VERSION_TEST
        )
        fs = os.path.getsize(archive_name)
        print("size: ", fs)
        self.assertGreater(fs, 0, "archive file size must be > 0")
        found = False
        for a in gh_release.get_assets():
            if a.name == archive_name:
                found = True
                self.assertEqual(fs, a.size, "asset size doesn't march archive size.")
                break
        self.assertTrue(found, "asset not found")

    def test_release_changelog(self):
        """Test if changelog in metadata.txt inside zipped plugin after release command."""
        # variables
        cli_config_changelog = Path("test/fixtures/.qgis-plugin-ci-test-changelog.yaml")
        version_to_release = "0.1.2"

        # load specific parameters
        with cli_config_changelog.open() as in_cfg:
            arg_dict = yaml.safe_load(in_cfg)
        parameters = Parameters(arg_dict)
        self.assertIsInstance(parameters, Parameters)

        # get output zip path
        archive_name = parameters.archive_name(
            plugin_name=parameters.plugin_path, release_version=version_to_release
        )

        # extract last items from changelog
        parser = ChangelogParser()
        self.assertTrue(parser.has_changelog())
        changelog_lastitems = parser.last_items(
            count=parameters.changelog_number_of_entries
        )

        # Include a changelog
        release(
            parameters=parameters,
            release_version=version_to_release,
            allow_uncommitted_changes=True,
        )

        # open archive and compare
        with ZipFile(archive_name, "r") as zip_file:
            data = zip_file.read(f"{parameters.plugin_path}/metadata.txt")

        # Changelog
        self.assertGreater(
            data.find(bytes(changelog_lastitems, "utf8")),
            0,
            f"changelog detection failed in release: {data}",
        )

        # Commit number
        self.assertEqual(1, len(re.findall(r"commitNumber=\d+", str(data))))

        # Commit sha1 not in the metadata.txt
        self.assertEqual(0, len(re.findall(r"commitSha1=\d+", str(data))))

    def test_release_version_valid_invalid(self):
        valid_tags = ["v1.1.1", "v1.1", "1.0.1", "1.1", "1.0.0-alpha", "1.0.0-dev"]
        invalid_tags = ["1", "v1", ".", ".1"]
        expected_valid_results = {
            "v1.1.1": ["v3"],
            "v1.1": ["v2"],
            "1.0.1": ["double", "semver"],
            "1.1": ["simple"],
            "1.0.0-alpha": ["semver"],
            "1.0.0-dev": ["semver"],
        }
        valid_results = {tag: [] for tag in valid_tags}
        patterns = Parameters.get_release_version_patterns()
        for key, cand in product(patterns, valid_results):
            if re.match(patterns[key], cand):
                valid_results[cand].append(key)
        self.assertEqual(valid_results, expected_valid_results)

        invalid_results = {tag: [] for tag in invalid_tags}
        for key, cand in product(patterns, invalid_results):
            if re.match(patterns[key], cand):
                invalid_results[cand].append(key)
        self.assertFalse(any(invalid_results.values()))

    def test_release_version_validation_on(self):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(
            title="commands", description="qgis-plugin-ci command", dest="command"
        )
        sub_parser = subparsers.add_parser("package")
        sub_parser.add_argument("release_version")
        sub_parser.add_argument("--no-validation", action="store_true")
        args = parser.parse_args(["package", "v1"])
        with self.assertRaises(ValueError):
            Parameters.validate_args(args)

    def test_release_version_validation_off(self):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(
            title="commands", description="qgis-plugin-ci command", dest="command"
        )
        sub_parser = subparsers.add_parser("package")
        sub_parser.add_argument("release_version")
        sub_parser.add_argument("--no-validation", action="store_true")
        args = parser.parse_args(["package", "v1", "--no-validation"])
        Parameters.validate_args(args)


if __name__ == "__main__":
    unittest.main()



#! python3  # noqa E265

"""
    Changelog parser. Following <https://keepachangelog.com/en/1.0.0/>.
"""

# ############################################################################
# ########## Libraries #############
# ##################################

# standard library
import logging
import re
import sys
from pathlib import Path
from typing import Union

from qgispluginci.version_note import VersionNote

# ############################################################################
# ########## Globals #############
# ################################

# see: https://regex101.com/r/8JROUv/1
CHANGELOG_REGEXP = r"(?<=##)\s*\[*(v?0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)\]?(\(.*\))?(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?\]*\s-\s*([\d\-/]{10})(.*?)(?=##|\Z)"
logger = logging.getLogger(__name__)

# ############################################################################
# ########## Classes #############
# ################################


class ChangelogParser:
    CHANGELOG_FILEPATH: Union[Path, None] = None

    @classmethod
    def has_changelog(
        cls, parent_folder: Union[Path, str] = Path("."), changelog_path="CHANGELOG.md"
    ) -> bool:
        """Check if a changelog file exists within the parent folder. If it does, \
        it returns True and the file path is stored as class attribute. If not, it \
        returns False and the class attribute is reset to None.

        Args:
            parent_folder (Union[Path, str], optional): parent folder where to look \
                for a `CHANGELOG.md` file. Defaults to Path(".").

            changelog_path str: Path relative to parent_folder. Defaults to CHANGELOG.md.

        Raises:
            FileExistsError: if the parent_folder path doesn't exist
            TypeError: if the path is not a folder but a path

        Returns:
            bool: True if a CHANGELOG.md exists within the parent_folder
        """
        # reset stored path as class attribute
        cls.CHANGELOG_FILEPATH = None

        # ensure using pathlib.Path
        if isinstance(parent_folder, str):
            parent_folder = Path(parent_folder)

        # check if the folder exists
        if not parent_folder.exists():
            logger.error(
                f"Parent folder doesn't exist: {parent_folder.resolve()}",
                exc_info=FileExistsError(),
            )
            sys.exit(1)
        # check if path is a folder
        if not parent_folder.is_dir():
            logger.error(
                f"Path is not a folder: {parent_folder.resolve()}", exc_info=TypeError()
            )
            sys.exit(1)

        # build, check and store the changelog path
        cls.CHANGELOG_FILEPATH = parent_folder / changelog_path
        if cls.CHANGELOG_FILEPATH.is_file():
            logger.info(f"Changelog file used: {cls.CHANGELOG_FILEPATH.resolve()}")
            return True
        else:
            logger.warning(
                f"Changelog file doesn't exist: {cls.CHANGELOG_FILEPATH.resolve()}"
            )
            cls.CHANGELOG_FILEPATH = None
            return False

    def __init__(
        self,
        parent_folder: Union[Path, str] = Path("."),
        changelog_path: str = "CHANGELOG.md",
    ):
        self.has_changelog(parent_folder=parent_folder, changelog_path=changelog_path)

    def _parse(self):
        if not self.CHANGELOG_FILEPATH:
            return None

        with self.CHANGELOG_FILEPATH.open(mode="r", encoding="UTF8") as f:
            content = f.read()

        return re.findall(
            pattern=CHANGELOG_REGEXP, string=content, flags=re.MULTILINE | re.DOTALL
        )

    def last_items(self, count: int) -> str:
        """Content to add in the metadata.txt.

        Args:
            count (int): Maximum number of tags to include in the file.

        Returns:
            str: changelog extraction ready to be added to metadata.txt
        """
        changelog_content = self._parse()
        if not changelog_content:
            return ""

        count = int(count)
        output = "\n"

        for version in changelog_content[0:count]:
            version_note = VersionNote(*version)
            output += f" Version {version_note.version}:\n"
            for item in version_note.text.split("\n"):
                if item:
                    output += f" {item}\n"
            output += "\n"
        return output

    def _version_note(self, tag: str) -> Union[VersionNote, None]:
        """Get the tuple for a given version."""
        changelog_content = self._parse()
        if not len(changelog_content):
            logger.error(
                f"Parsing the changelog ({self.CHANGELOG_FILEPATH.resolve()}) "
                "returned an empty content."
            )
            return None

        if tag == "latest":
            return VersionNote(*changelog_content[0])

        for version in changelog_content:
            version_note = VersionNote(*version)
            if version_note.version == tag:
                return version_note

    def latest_version(self) -> str:
        """Return the latest tag described in the changelog file."""
        latest = self._version_note("latest")
        logger.debug(
            "Latest version retrieved from changelog "
            f"({self.CHANGELOG_FILEPATH.resolve()}): {latest.version}"
        )
        return latest.version

    def content(self, tag: str) -> Union[str, None]:
        """Get a version content to add in a release according to the version name."""
        version_note = self._version_note(tag)
        if not version_note:
            return None

        return version_note.text


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    pass



#!/usr/bin/env python
"""
/***************************************************************************

 QGIS Plugin CI Testing
 Copyright (C) 2019 Denis Rouzaud

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

DEBUG = False

import os

from PyQt5.QtCore import QCoreApplication, QLocale, QSettings, QTranslator
from qgis.gui import QgisInterface


class QgisPluginCiTesting:
    def __init__(self, iface: QgisInterface):
        self.iface = iface

        # initialize translation
        qgis_locale = QLocale(QSettings().value("locale/userLocale"))
        locale_path = os.path.join(os.path.dirname(__file__), "i18n")
        self.translator = QTranslator()
        self.translator.load(qgis_locale, "swiss_locator", "_", locale_path)
        QCoreApplication.installTranslator(self.translator)

        self.trUtf8("some UTF-8 translation: un épilogue où l'on marche sur des œufs")

    def initGui(self):
        pass

    def unload(self):
        pass
